__productname__ = "analytix"
__version__ = "2.0.0.dev5"
__description__ = "A simple yet powerful API wrapper to make getting analytical information from the YouTube Analytics API easier than ever."
__url__ = "https://github.com/parafoxia/analytix"
__docs__ = "https://analytix.readthedocs.io/en/latest/"
__author__ = "Ethan Henderson"
__license__ = "BSD-3-Clause"
__bugtracker__ = "https://github.com/parafoxia/analytix/issues"

from .youtube.analytics.api import YouTubeAnalytics
